#############
# $Id: Connected.pm 12048 2010-12-02 12:30:28Z gq1 $
# Created By: ajb
# Mast Maintained By: $Author: gq1 $
# Created On: 2009-06-01
# Last Changed On: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/Net/Stomp/Receipt/Connected.pm $
package Net::Stomp::Receipt::Connected;

use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};

#
# Subclass of Net::Stomp::Receipt for testing the socket just prior to sending message to socket
# with receipt and commit.
#
# Author: Hugo Salgado <huguei@cpan.org>
#
use base 'Net::Stomp::Receipt';

#
# I start the version in sync with Net::Stomp
# (and hopes to keep it in that way)
use Readonly; Readonly::Scalar our $VERSION => '0.35';


# Added a new configuration variable on creation,
# so we need to subclass the constructor

# The send_safe method overwritten to include testing for connectedness before any calls to send_frame
sub send_safe {
    my ( $self, $conf ) = @_;
    eval { $self->socket->connected(); } or do { croak 'Message send aborted - Error with socket connection.'};
    my $body = $conf->{body};
    delete $conf->{body};

    # Transaction begins
    $conf->{transaction} = $self->{'sessionid'} .  q{-} .  $self->_getSerial;
    my $frame = Net::Stomp::Frame->new(
        { command => 'BEGIN', headers => $conf } );
    $self->send_frame($frame);
    undef $frame;

    # Sending the message, with receipt header
    my $receipt_id = $self->{'sessionid'} .  q{-} .  $self->_getSerial;
    my $return;
    eval {
      $conf->{receipt} = $receipt_id;
      if ($self->{'PERSISTENT'}) {
        $conf->{persistent} = 'true' ;
      }
      $frame = Net::Stomp::Frame->new(
        { command => 'SEND', headers => $conf, body => $body } );
      eval { $self->socket->connected(); } or do { croak 'Error with socket connection before message send.'};
      $self->send_frame($frame);
      undef $frame;
      delete $conf->{receipt};
      delete $conf->{persistent};

      # Checking the server for the right receipt
      # If it's OK -> commit the transaction
      $frame = $self->receive_frame;
      if (($frame->command eq 'RECEIPT') and
        ($frame->headers->{'receipt-id'} eq $receipt_id)) {
        my $frame_commit = Net::Stomp::Frame->new(
            { command => 'COMMIT', headers => $conf } );
        eval { $self->socket->connected(); } or do { croak 'Error with socket connection before commit send.'};
        $self->send_frame($frame_commit);

        $return = 1;
      }
      1;
    } or do {
      carp "Error with connection: $EVAL_ERROR";
    };

    if (!$return) {
      # whatever else, abort transaction
      my $frame_abort = Net::Stomp::Frame->new(
        { command => 'ABORT', headers => $conf } );
      eval { $self->socket->connected(); } or do { croak 'Error with socket connection before abort send.'};
      $self->send_frame($frame_abort);

      croak qq{Message Send Aborted $EVAL_ERROR};
    }

    return $return;
}

# The send_frame method overwritten to add a test for connectedness before the print to the socket.
sub send_frame {
    my ( $self, $frame ) = @_;

#         warn "send [" . $frame->as_string . "]\n";
    my $sock = $self->socket();
    $sock->connected() or croak q{Aborted sending_frame, as cannot ping connection};
    my $return = $sock->print( $frame->as_string() ) or croak $EVAL_ERROR;
    return $return;
}

1;

__END__

=head1 NAME

Net::Stomp::Receipt::Connected - An extension to Net::Stomp::Receipt which extends the transactional sends to include testing the connection.

=head1 VERSION

  0.35

=head1 SYNOPSIS

  use Net::Stomp::Receipt::Connected;

  my $stomp = Net::Stomp::Receipt::Connected->new({
    hostname   => 'localhost',
    port       => '61613',
    PERSISTENT => 1,
  });

  $stomp->connect({
    login    => 'hello',
    passcode => 'there'
  });

  $stomp->send_safe({
    destination => '/queue/foo',
    body => 'test message'
  }) or die "Couldn't send the message!";

=head1 SUBROUTINES/METHODS

=head2 connect

=head2 send_safe

=head2 send_frame

=head1 DESCRIPTION

This module is an extension to Net::Stomp module, an Streaming
Text Orientated Messaging Protocol client, that adds a new
method send_safe which uses "transactional sends".

By this way, any message sent to the stomp server is identified
with a transaction id, that must be acked by a server receipt.
In case of failure, the send is aborted.

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Net::Stomp::Receipt

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 SEE ALSO

Net::Stomp module

The protocol spec: http://stomp.codehaus.org/Protocol

=head1 AUTHOR

Hugo Salgado, E<lt>huguei@cpan.org<gt>

Extended by

Andy Brown <ajb@sanger.ac.uk>


=head1 ACKNOWLEDGEMENTS

This module was built for NIC Chile (http://www.nic.cl), who
granted its liberation as free software.

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2008 by Hugo Salgado
Copyright (C) 2009 by Andy Brown

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.10.0 or,
at your option, any later version of Perl 5 you may have available.

=cut
