#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2010 01 15
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: Loader.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/irods/Loader.pm $
#

package npg_common::irods::Loader;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Basename;
use File::Spec;
use IPC::Open3;
use Digest::MD5 qw(md5);
use Perl6::Slurp;

with 'MooseX::Getopt';
with 'npg_common::roles::log';

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/mxs; $r; };
Readonly::Scalar our $ICOMMAND_PATH   => q{/software/irods/icommands/bin/};
Readonly::Scalar our $EXIT_CODE_SHIFT => 8;

## no critic (Documentation::RequirePodAtEnd ProhibitBacktickOperators)

=head1 NAME

npg_common::irods::Loader

=head1 VERSION

$LastChangedRevision: 12048 $

=head1 SYNOPSIS

    my $loader = npg_common::irods::Loader->new({
       file         => $file_to_load,
       resource     => $destination_resourse,
       collection   => $destination_dir,
       new_filename => $destination_filename
       meta_data    => $meta_hashref,
    });

=head1 DESCRIPTION

This module load a file to irods server

you must run iinit command first

=head1 SUBROUTINES/METHODS

=head2 file

file to be uploaded, with full path

=cut

has 'file'        => (isa           => 'Str',
                      is            => 'rw',
                      required      => 1,
                      documentation => 'file to be uploaded, with path',
                     );
=head2 resource

which resource should the file be loaded to, otherwise use default

=cut
has 'resource'   =>  (isa           => 'Str',
                      is            => 'rw',
                      required      => 0,
                      documentation => 'resource to use',
                     );

=head2 collection

which collection to load the file, otherwise current

=cut
has 'collection'  => (isa           => 'Str',
                      is            => 'rw',
                      required      => 0,
                      documentation => 'collection to load the file',
                     );

=head2 new_filename

given a new filename if you want to change it

=cut
has 'new_filename' =>(isa           => 'Str',
                      is            => 'rw',
                      required      => 0,
                      lazy_build    => 1,
                      documentation => 'new filename if you need change',
                     );

sub _build_new_filename{
   my $self = shift;
   return basename($self->file());
}

=head2 chmod_permission

chmod permission string, in the following format:
null|read|write|own userOrGroup, for example,
'null public' will give no any permission for public group or user

=cut

has 'chmod_permission' => (isa           => 'Str',
                           is            => 'rw',
                           required      => 0,
                           documentation => 'chmod permission string',
                         );

=head2 meta_data

a list of meta data as hash ref

=cut
has 'meta_data'   => (isa           => 'HashRef',
                      is            => 'rw',
                      required      => 0,
                      default       => sub { {} },
                      documentation => 'meta data to add',
                     );

=head2 run

main method to call, load the file and add meta data 

=cut
sub run{
  my $self = shift;

  my $des_file = $self->new_filename;

  my $des_dir = $self->collection();

  if($des_dir){

    $self->log('Making destination directory when neccesary');
    my $mkdir_cmd = $ICOMMAND_PATH.qq{imkdir -p $des_dir};
    $self->_run_cmd($mkdir_cmd);

    $des_file = File::Spec->catfile($des_dir, $des_file);
  }

  $self->log('Checking md5 of the file');
  my $local_md5 = $self->_get_file_md5($self->file());
  my $remote_md5 = $self->_get_irods_md5($des_file);

  if( defined $remote_md5 && ( $remote_md5 eq $local_md5 ) ){
     $self->log( 'FILE EXISTS! NO NEED TO LOAD: ' . $self->file );
     $self->add_meta($des_file, $self->meta_data());
     return;
  }

  my $iput_cmd = $ICOMMAND_PATH. q{iput -K};

  if($self->resource()){

     $iput_cmd .=q{ -R }.$self->resource();
  }

  $iput_cmd .= q{ -f};

  $iput_cmd .= q{ }.$self->file().q{ }.$des_file;

  $self->_run_cmd($iput_cmd);

  if( $self->chmod_permission() ){
     my $ichmod_cmd =  $ICOMMAND_PATH. q{ichmod }. $self->chmod_permission().q{ }.$des_file;
     $self->_run_cmd($ichmod_cmd );
  }

  $self->add_meta($des_file, $self->meta_data());

  return 1;

}

=head2 add_meta

given a file on the server and a list of meta data, add the meta data to the file

=cut
sub add_meta{
  my ($self, $file, $meta_data) = @_;

  my $remote_meta_data = $self->_check_meta_data($file);

  foreach my $att_name (keys %{$meta_data}){
     my $values = $meta_data->{$att_name};
     my @add_meta_cmd = ( $ICOMMAND_PATH.q{imeta}, q{add}, q{-d}, $file, $att_name );

     if( ref $values eq q{ARRAY} ){
       foreach my $value (@{$values}){
         if( ! $remote_meta_data->{$att_name}->{$value} ){
            push @add_meta_cmd, $value;
            $self->_run_cmd (@add_meta_cmd);
            pop  @add_meta_cmd;
         }
       }
     }else{
       if( ! $remote_meta_data->{$att_name}->{$values} ){
        push @add_meta_cmd, $values;
        $self->_run_cmd (@add_meta_cmd);
        pop  @add_meta_cmd;
       }
     }

  }
  return 1;
}

sub _run_cmd {
  my ($self, @cmd) = @_;
  $self->log(join q{ }, @cmd);
  my $output_meta = system @cmd;
  if($CHILD_ERROR){
     croak q{Failed: }.(join q{ }, @cmd).qq{\n$CHILD_ERROR $output_meta};
  }
  return;
}

sub _check_meta_data {
  my ($self, $des_file) = @_;

  my $imeta_ls_cmd = $ICOMMAND_PATH . qq{imeta ls -d $des_file};
  $self->log("Checking meta data: $imeta_ls_cmd");

  my $pid = open3( undef, my $imeta_ls_out_fh, undef, $imeta_ls_cmd);

  my ($name, $value);
  my %meta_data = ();

  while (my $line = <$imeta_ls_out_fh> ) {

    chomp $line;
    #$self->log($line);
    if( $line =~ /^[-]{4}/mxs ){
       $name = undef;
       $value = undef;
    }elsif ( $line =~ /^attribute[:][ ](.+)/mxs ){
      $name = $1;
    }elsif( $line =~ /^value[:][ ](.+)/mxs ) {
      $value = $1;
    }

    if(defined $name && defined $value) {
       $meta_data{$name}->{$value}++;
    }
  }

  waitpid $pid, 0;

  if( $CHILD_ERROR >> $EXIT_CODE_SHIFT){
     croak "Failed: $imeta_ls_cmd";
  }

  close $imeta_ls_out_fh or croak "can not close imeta ls command output: $ERRNO";

  return \%meta_data;
}

sub _get_irods_md5 {
  my ($self, $des_file) = @_;

  my %des_file_md5_hash = ();

  my $ils_cmd = $ICOMMAND_PATH . qq{ils -L $des_file};
  $self->log($ils_cmd);

  my $pid = open3( undef, my $ils_out_fh, undef, $ils_cmd);

  while (my $line = <$ils_out_fh> ) {

    chomp $line;

    if ( $line =~ /ERROR[:][ ]lsUtil[:][ ]/mxs ){
       #file not exists in the server
       last;
    }elsif ($line =~ /^[ ]{4}(\w+)[ ]/mxs){
       #file exists in the server and md5 available
       $des_file_md5_hash{$1}++;
    }
  }

  waitpid $pid, 0;

  if( $CHILD_ERROR >> $EXIT_CODE_SHIFT ){
     carp "$des_file not exist on irods";
  }

  close $ils_out_fh or croak "can not close ils command output: $ERRNO";

  my $num_md5_values = scalar keys %des_file_md5_hash;

  if( $num_md5_values > 1){
     #different md5 on different servers
     return 0;
  }elsif( $num_md5_values == 0 ){
     #file not exists or md5 not available
     return ;
  }

  my @md5_values = keys %des_file_md5_hash;
  return shift @md5_values;
}

sub _get_file_md5 {
  my ($self, $file) = @_;

  my $md5_file = $file.q{.md5};
  if(-e $md5_file){
    my @md5_lines = slurp $md5_file;
    my $md5_from_file = shift @md5_lines;
    if($md5_from_file){
       return $md5_from_file;
    }
  }

  open my $fh, '<', $file or croak "Can't open '$file': $ERRNO";
  binmode $fh;

  my $md5 = Digest::MD5->new->addfile($fh)->hexdigest;
  close $fh or croak "Can't close '$file': $ERRNO";

  return $md5;
}

=head2 get_collection_file_list

given a collection, return a hashref with a list of files in this collection

=cut

sub get_collection_file_list {
  my ( $self, $collection) = @_;

  my %file_list = ();

  my $ils_cmd = $ICOMMAND_PATH . qq{ils $collection};
  $self->log($ils_cmd);

  my $pid = open3( undef, my $ils_out_fh, undef, $ils_cmd);

  while (my $line = <$ils_out_fh> ) {

    chomp $line;

    if ( $line =~ /ERROR[:][ ]lsUtil[:][ ]/mxs ){
       #collection not exists in the server
       last;
    }elsif ($line !~ /$collection/mxs){
       #chomp $line;
       $line =~ s/\s+//mxs;
       $file_list{$line}++;
    }
  }

  waitpid $pid, 0;

  if( $CHILD_ERROR >> $EXIT_CODE_SHIFT ){
     croak "Failed: $ils_cmd";
  }

  close $ils_out_fh or croak "can not close ils command output: $ERRNO";

  return \%file_list;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item  File::Basename

=item  File::Spec

=item  npg_common::roles::log

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
