#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2010 01 15
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: Bam.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/irods/run/Bam.pm $
#

package npg_common::irods::run::Bam;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Basename;
use File::Spec;
use npg::api::run;
use npg::api::run_lane;
use st::api::asset;
use npg_common::irods::Loader;
use List::Util qw(first);
use IPC::Open3;

with qw{MooseX::Getopt npg_common::roles::run::short_info npg_common::roles::run::path_info npg_common::roles::log};
with qw{npg_common::roles::run::long_info};
with qw{npg_common::roles::run::lane::tag_info};

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $DEFAULT_RESOURCE   => q{seq};
Readonly::Scalar our $DEFAULT_ROOT_DIR   => q{/seq/};
Readonly::Scalar our $DEFAULT_LANE_NUMBERS   => 8;
Readonly::Scalar our $EXIT_CODE_SHIFT => 8;
Readonly::Scalar our $SAMTOOLS => q{/software/solexa/bin/aligners/samtools/current/samtools};

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::irods::run::Bam

=head1 VERSION

$LastChangedRevision: 12048 $

=head1 SYNOPSIS


=head1 DESCRIPTION


=head1 SUBROUTINES/METHODS

=head2 positions

a list of postions of the run

=cut
has 'positions'    => (isa           => q{ArrayRef},
                       is            => q{rw},
                       documentation => q{a list of postions of the run},
                      );

=head2 npg_api_run

npg api run object

=cut
has 'npg_api_run'  => (isa           => q{npg::api::run},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{npg api run object},
                      );

sub _build_npg_api_run {
   my $self = shift;
   return npg::api::run->new( {id_run => $self->id_run() } );
}

sub _build_run_folder {
    my ($self) = @_;
    if (! ($self->_given_path or $self->has_id_run or $self->has_name)){
       croak 'need a path to work out a run_folder';
    }
    return first {$_ ne q()} reverse File::Spec->splitdir($self->runfolder_path);
}

=head2 file_list

bam file list
 
=cut
has 'file_list'    => (isa           => q{ArrayRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{bam file name list},
                      );

sub _build_file_list {
  my $self = shift;

  my @file_list = ();

  my $positions = $self->positions();
  my $positions_string = q{*};
  if( $positions && scalar @{$positions} > 0 ) {
    $positions_string = join q{,}, @{$positions};
    $positions_string = qq{{$positions_string}};
  }

  @file_list = glob  $self->archive_path().q{/}.$self->id_run.q{_}.$positions_string.q{*.bam};

  my @file_list_plex = glob $self->archive_path().q{/lane}.$positions_string.q{/*.bam};

  push @file_list, @file_list_plex;

  return \@file_list;
}

=head2 exclude_bam_index

a flag to include bam index file in the loading,
The index bam file must be in the same directory as bam file, 
and the file name ends with .bai

=cut

has 'exclude_bam_index'   => (isa           => q{Bool},
                              is            => q{rw},
                              documentation => q{flag to exclue bam index file loading},
                             );

=head2 process

main method to call, get each bam file, load them and add meta data 
 
=cut

sub process{ ## no critic (Subroutines::ProhibitExcessComplexit)
  my $self = shift;

  my $id_run = $self->id_run();

  $self->log("Loading bam file for run $id_run");
  if( $self->positions() && ( scalar @{$self->positions()} ) ){
    $self->log( 'Lane '. (join q{, }, @{$self->positions}) );
  }

  my $file_list = $self->file_list();
  $self->log('There are '.( scalar @{ $file_list} ). ' bam files to load');

  my $file_count = 0;
  foreach my $file (@{ $file_list}){

    $self->log("---- Loading file $file");

    my $file_name = basename($file);
    my ($lane) = $file_name =~ /\d+_(\d)/mxs;
    my ($tag_index) = $file_name =~ /\#(\d+)\.bam/mxs;
    my ($human_split) = $file_name =~ /\d+_\d_(\w+)/mxs;

    if( (!defined $tag_index) && $self->is_multiplexed_lane($lane)){
       $self->log('Multiplex lane bam file, IGNORED');
       next;
    }

    my $names = $self->get_study_library_sample_names($id_run, $lane, $tag_index);

    my $study_names = $names->{study};
    my $library_names = $names->{library};
    my $sample_names = $names->{sample};
    my $is_phix_control =  $names->{is_phix_control};
    my $is_human_split =  $names->{is_human_split};

    if( $is_phix_control ) {
       $self->log('Phix control lane, IGNORED');
       next;
    }

    if( (!$is_human_split) && $human_split ){
       $self->log('Human split bam file for study without unconsented data, IGNORED');
       next;
    }

    if( $is_human_split && (!$human_split) ){
       $self->log('Original bam file for study with unconsented data, IGNORED');
       next;
    }

    my $meta  = {
                  id_run => $id_run,
                  lane   => $lane,
                  type   => q{bam},
                };

    if(defined $tag_index){
       $meta->{tag_index} = $tag_index;

       my $tag_list = $self->tag_list($lane);
       my $tag = $tag_list->{$tag_index};
       if($tag){
         $meta->{tag} = $tag;
       }
    }

    if($human_split){
       $meta->{human_split} = $human_split;
    }

    if(scalar @{$study_names}){
      $meta->{study} = $study_names;
    }
    if(scalar @{$library_names}){
      $meta->{library} = $library_names;
    }
    if(scalar @{$sample_names}){
      $meta->{sample} = $sample_names;
    }

    my $alignment = 0;
    my $reference = $self->get_bam_reference($file);
    if( $reference ){
       $meta->{reference} = $reference;
       $alignment = 1;
    }
    $meta->{alignment} = $alignment;

    my $loader = npg_common::irods::Loader->new({
       file        => $file,
       resource    => $DEFAULT_RESOURCE,
       collection  => $DEFAULT_ROOT_DIR . $id_run,
       meta_data   => $meta,
    });

    if( $human_split && $human_split eq q{human} ){
       #give null permission for public group if bam is human part
       $loader->chmod_permission(q{null public});
    }

    if($loader->run()){
       $file_count++;
    }

    $self->_load_bam_index_file($file, $id_run, $human_split);

  }

  $self->log("---- $file_count bam files loaded for run $id_run");

  if(!$self->bam_fully_archived){
     croak 'BAM files not fully archived';
  }else{
     $self->log('BAM files are fully archived');
  }

  return 1;
}

sub _load_bam_index_file {
    my ($self, $bam_file, $id_run, $human_split) = @_;

    if( $self->exclude_bam_index() ){
       return;
    }

    my $index_file = $bam_file;
    $index_file =~ s/bam$/bai/mxs;

    if( !-e $index_file ){
       croak "Bam index file not exist: $index_file";
    }

    $self->log("-Loading bam index file: $index_file");
    my $loader = npg_common::irods::Loader->new({
       file        => $index_file,
       resource    => $DEFAULT_RESOURCE,
       collection  => $DEFAULT_ROOT_DIR . $id_run,
       meta_data   => {type => q{bai},},
    });

    if( $human_split && $human_split eq q{human} ){
       #give null permission for public group if bam is human part
       $loader->chmod_permission(q{null public});
    }

    $loader->run();

    return 1;
}

=head2 get_study_library_sample_names

method to get study, library and sample names for a bam file
 
=cut

sub get_study_library_sample_names {
   my ($self, $id_run, $position, $tag_index) = @_;

   my $study_names = [];
   my $library_names = [];
   my $sample_names = [];
   my $is_phix_control = 0;
   my $is_human_split = 0;

   my $asset;

   my $run_lane = npg::api::run_lane->new( {id_run => $id_run, position => $position} );

   if ( $run_lane->pool && $tag_index ) {
      my $st_lane = $run_lane->st_lane();
      if (!$st_lane->pool_description) {
         croak qq[No pool description for run $id_run position $position];
      }
      if (exists $st_lane->pool_description->{$tag_index}->{lib_tube_asset_id}) {
        $asset = st::api::asset->new( {id => $st_lane->pool_description->{$tag_index}->{lib_tube_asset_id}} );
      } else {
        carp qq[Library tube asset for run $id_run position $position tag index $tag_index not found];
      }
   } else {
     $asset = $run_lane->entity;
   }

   my $studies;
   my $samples;
   my $asset_name;

   if($asset){
     $studies = $asset->studies();
     $samples = $asset->ancestor_samples();
     $asset_name = $asset->name();
   }

   if( $studies ){
     foreach my $study (@{$studies}){
        push @{$study_names}, $study->name;
        if($study->contains_unconsented_human()){
           $is_human_split = 1;
        }
     }
   }

   if($asset_name ){
     push @{ $library_names }, $asset_name;
   }

   if( $samples ){
     foreach my $sample (@{$samples}){
        push @{$sample_names}, $sample->name;
     }
   }

   if( $run_lane->control() && $asset_name =~ /phix/ixms ){
      $is_phix_control = 1;
   }

   my $names = { study           => $study_names,
                 library         => $library_names,
                 sample          => $sample_names,
                 is_phix_control => $is_phix_control,
                 is_human_split  => $is_human_split,
               };

  return $names;
}

=head2 bam_fully_archived_for_deletion

simply checking bam file in archive for deletion
 
=cut

sub bam_fully_archived_for_deletion { ## no critic (Subroutines::ProhibitExcessComplexit)
    my $self = shift;

    my $fully_archived = 1;

    my $bam_list_archived;
    eval{
        $bam_list_archived = npg_common::irods::Loader->new(file=>'none')
                            ->get_collection_file_list($DEFAULT_ROOT_DIR.$self->id_run());
	     1;
    } or do {
        carp "Problems to check irods: $EVAL_ERROR";
	     return 0;
    };

    my @file_list = keys %{$bam_list_archived};
    my %bam_list = map {$_=> 1 } grep { /bam$/mxs } @file_list;
    my %bai_list = map {$_=> 1 } grep { /bai$/mxs } @file_list;

    if(! $self->_check_bam_index_file(\%bam_list, \%bai_list)){
       $fully_archived = 0;
    }

    my %bam_list_by_lane = ();
    my %bam_list_by_lane_split = ();
    my %bam_list_by_lane_tag = ();
    my %bam_list_by_lane_split_tag = ();

    foreach my $file_name (keys %bam_list){

        my ($lane) = $file_name =~ /\d+_(\d)/mxs;
        my ($tag_index) = $file_name =~ /\#(\d+)\.bam/mxs;
        my ($human_split) = $file_name =~ /\d+_\d_(\w+)/mxs;

        if(!$lane){
           croak "There is no lane number in bam file $file_name";
        }

        if ( !$human_split ){
           if (!defined $tag_index){
             $bam_list_by_lane{$lane}++;
           } else{
             $bam_list_by_lane_tag{$lane}->{$tag_index}++;
           }
        } elsif ( $human_split) {
           if (!defined $tag_index){
             $bam_list_by_lane_split{$lane}->{human_split}++;
           } else{
             $bam_list_by_lane_split_tag{$lane}->{human_split}->{$tag_index}++;
           }
        }
    }

    my $positions = $self->positions();
    if( ! $self->positions() || ( scalar @{$self->positions()} == 0) ){
       $positions = [1..$DEFAULT_LANE_NUMBERS];
    }

    foreach my $position (@{$positions}){

          my $st_info_lane = $self->get_study_library_sample_names($self->id_run(), $position);
          my $is_phix_control_lane = $st_info_lane ->{is_phix_control};
          my $is_human_split_lane = $st_info_lane->{is_human_split};

          my $is_multiplexed = $self->is_multiplexed_lane($position);

          if($is_phix_control_lane){
             next;
          }elsif(!$is_multiplexed){
             my $number_bam_files = $bam_list_by_lane_split{$position}->{human_split};
             if( ( $is_human_split_lane &&  ( !$number_bam_files || $number_bam_files < 2) )
                || (!$is_human_split_lane &&  !$bam_list_by_lane{$position} ) ){

                   $self->log("Lane $position bam file missing");
                   $fully_archived = 0;
              }
          }else {
              my $tag_index_list = $self->get_tag_index_list($position);

              foreach my $tag_index (@{$tag_index_list}){
                 if( !$bam_list_by_lane_tag{$position}->{$tag_index}
                     && !$bam_list_by_lane_split_tag{$position}->{human_split}->{$tag_index} ){
                         $self->log("Lane $position tag $tag_index bam file missing");
                         $fully_archived = 0;
                     }
              }

          }
    }

    return $fully_archived;
}

sub _check_bam_index_file {
    my ($self, $bam_list_ref, $bai_list_ref) = @_;

    if($self->exclude_bam_index()){
       return 1;
    }

    if(scalar keys %{$bam_list_ref} != scalar keys %{$bai_list_ref}){
       $self->log('The numbers of bam files and bai file not match');
       return 0;
    }

    my $index_file_found = 1;
    foreach my $bam ( keys %{$bam_list_ref} ){
       my $bai = $bam;
       $bai =~ s/bam$/bai/mxs;
       if(!$bai_list_ref->{$bai}){
          $self->log("Index file for $bam not exist");
          $index_file_found = 0;
       }
    }

    return $index_file_found;
}

=head2 bam_fully_archived

check every bam file fully archived
 
=cut

sub bam_fully_archived {
    my $self = shift;

    my $fully_archived = 1;

    my $bam_list_archived = npg_common::irods::Loader->new(file=>'none')
                            ->get_collection_file_list($DEFAULT_ROOT_DIR.$self->id_run());
    my $bam_list_to_archive = $self-> _get_bam_list_to_archive();

    foreach my $bam (@{$bam_list_to_archive}){
       if(!$bam_list_archived->{$bam}){
          $fully_archived = 0;
          $self->log("File Missing in iRods: $bam");
       }

       if( $self->exclude_bam_index() ){
           next;
       }

       my $bam_index_file = $bam;
       $bam_index_file =~ s/bam$/bai/mxs;
       if(!$bam_list_archived->{$bam_index_file}){
          $fully_archived = 0;
          $self->log("File Missing in iRods: $bam_index_file");
       }

    }

    return $fully_archived;
}

sub _get_bam_list_to_archive {
    my $self = shift;

    $self->log('Getting a list of bam files which should be archived');

    my @bam_list = ();

    my $positions = $self->positions();
    if( ! $self->positions() || ( scalar @{$self->positions()} == 0) ){
       $positions = [1..$DEFAULT_LANE_NUMBERS];
    }

    foreach my $position (@{$positions}){
       push @bam_list, @{$self->_bam_file_list_per_lane($position)};
    }
    return \@bam_list;
}

sub _bam_file_list_per_lane {
   my ($self, $position) = @_;

   my @bam_list_per_lane = ();

   my $st_info_lane = $self->get_study_library_sample_names($self->id_run(), $position);
   my $is_phix_control_lane = $st_info_lane ->{is_phix_control};
   my $is_human_split_lane = $st_info_lane->{is_human_split};

   if( !$self->is_multiplexed_lane($position) && !$is_phix_control_lane ){
          if( $is_human_split_lane ){
             push @bam_list_per_lane, $self->_bam_file_name({id_run      => $self->id_run(),
                                                             position    => $position,
                                                             human_split => 'human'});
             push @bam_list_per_lane, $self->_bam_file_name({id_run      => $self->id_run(),
                                                             position    => $position,
                                                             human_split => 'nonhuman'});
          }else{
             push @bam_list_per_lane, $self->_bam_file_name({id_run   => $self->id_run(),
                                                             position => $position});
          }
    }elsif(!$is_phix_control_lane){

          my $tag_index_list = $self->get_tag_index_list($position);

          foreach my $tag_index (@{$tag_index_list}){
             my $st_info_plex = $self->get_study_library_sample_names($self->id_run(), $position, $tag_index);
             my $is_human_split_plex = $st_info_plex->{is_human_split};

             if( $is_human_split_plex ){
                push @bam_list_per_lane, $self->_bam_file_name({id_run      => $self->id_run(),
                                                                position    => $position,
                                                                human_split => 'human',
                                                                tag_index   => $tag_index,});
                push @bam_list_per_lane, $self->_bam_file_name({id_run      => $self->id_run(),
                                                                position    => $position,
                                                                human_split => 'nonhuman',
                                                                tag_index   => $tag_index,});
             }else{
                push @bam_list_per_lane, $self->_bam_file_name({id_run   => $self->id_run(),
                                                                position => $position,
                                                                tag_index   => $tag_index,});
             }
          }
    }
    return \@bam_list_per_lane;
}

sub _bam_file_name {
   my ($self, $args_ref) = @_;

   my $id_run      = $args_ref->{id_run};
   my $position    = $args_ref->{position};
   my $human_split = $args_ref->{human_split};
   my $tag_index   = $args_ref->{tag_index};

   my $bam_file_name = $self->id_run.q{_}.$position;
   if( $human_split ){
      $bam_file_name .= q{_}.$human_split;
   }
   if( defined $tag_index ){
      $bam_file_name .= q{#}.$tag_index;
   }

   return $bam_file_name.q{.bam};
}

sub _get_ref_from_bwa_pg {
   my ($self, $bwa_pg_line) = @_;

   my @ref = grep {/\/nfs\/repository\/d0031\/references/mxs} split /\s/mxs, $bwa_pg_line;
   if(scalar @ref){
      return $ref[0];
   }
   return;
}

sub _grep_bam_bwa_pg_line {
   my ( $self, $sam_header_fh ) = @_;

   while( my $line = <$sam_header_fh> ) {
      if( $line =~ /^\@PG/mxs && $line =~/ID\:bwa/mxs && $line !~ /ID\:bwa_sam/mxs) {
          chomp $line;
          return $line;
      }
   }
   return;
}

=head2 get_bam_reference

read bam header using samtools and check bwa or bwa_aln PG to get the reference used
 
=cut

sub get_bam_reference {
   my ( $self, $bam ) = @_;

   my $samtools_view_cmd = qq{$SAMTOOLS view -H $bam};
   $self->log($samtools_view_cmd);

   my $pid = open3( undef, my $sam_header_fh, undef, $samtools_view_cmd);

   my $bwa_pg_line = $self->_grep_bam_bwa_pg_line($sam_header_fh);

   my $reference;
   if($bwa_pg_line){
      $reference = $self->_get_ref_from_bwa_pg( $bwa_pg_line );
   }
   waitpid $pid, 0;

   if( $CHILD_ERROR >> $EXIT_CODE_SHIFT ){
     croak qq{Failed: $samtools_view_cmd};
   }

   close $sam_header_fh or croak "can not close samtools view output: $ERRNO";

   if($reference){
     $self->log("Reference found from bam header: $reference");
   }else{
       $self->log('NO reference found from bam header.');
   }
   return $reference;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item File::Basename

=item File::Spec

=item npg::api::run

=item npg::api::run_lane

=item st::api::asset

=item npg_common::irods::Loader

=item npg_common::roles::run::short_info

=item npg_common::roles::run::path_info

=item npg_common::roles::log

=item npg_common::roles::run::long_info

=item npg_common::roles::run::lane::tag_info

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
