#########
# Author:        Andy Brown and Marina Gourtovaia
# Maintainer:    $Author: mg8 $
# Created:       03 April 2009
# Last Modified: $Date: 2009-04-08 18:40:03 +0100 (Wed, 08 Apr 2009) $
# Id:            $Id: qXvalues.pm 4831 2009-04-08 17:40:03Z mg8 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/trunk/lib/npg_common/qXvalues.pm $
#

package npg_common::qXvalues;

use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};
use Exporter;
use Math::Round qw(round);

## no critic (ProhibitExplicitISA RequireLocalizedPunctuationVars)

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12048 $ =~ /(\d+)/mxs; $r; };
our @ISA     = qw(Exporter);
our @EXPORT_OK  = qw(get_qvals check_int);

Readonly our $POSITION_OF_QVAL_LOWEST     => 26;
Readonly our $LOWEST_VALUE                => 20;
Readonly our $HIGHEST_VALUE               => 42;
Readonly our $COEFF                       => 1000;
Readonly our $NUM_OTHER_FIELDS            => 6;
Readonly our $TOP_ROW_NUM_FIELDS          => 9;
Readonly our $MIN_NUM_WORDS               => 5;

## no critic (ProhibitExcessComplexity)

our $DOWARN = 1;

BEGIN { $SIG{'__WARN__'} = sub { croak $_[0] if !$DOWARN; carp $_[0] } }

sub check_int {

    my $value = shift;
    my $int_value = 0;
    $DOWARN = 0;
    ## no critic (RequireCheckingReturnValueOfEval)
    eval {
        $int_value = int $value;
    };
    $DOWARN = 1;
    if (!$EVAL_ERROR && $int_value == $value) {
        return 1;
    }
    return 0;
}

sub _has_zero_sequences {
    my $words = shift;
    my $result = 0;
    if(@{$words} >= $MIN_NUM_WORDS && $words->[0] == 0 && $words->[2] == 0)  {$result = 1;}
    return $result;
}

sub get_qvals {
    my ($file_path, $thresholds) = @_;

    if (@_ != 2) {
        croak q[Wrong number of arguments to ] . __PACKAGE__ . q[::get_qvals];
    }

    if (ref($thresholds) ne q[ARRAY]) {
        croak q[Wrong type for the second argument in ] .  __PACKAGE__ . q[::get_qvals];
    }

    carp q[This module is depricated. Please use npg_common::fastqcheck module.];

    open my $fh, q[<], $file_path or croak "Cannot open $file_path";

    my @results = ();

    # read and parse first line
    my $line = <$fh>;
    my @temp = split /\s+/xms, $line;
    if (@temp != $TOP_ROW_NUM_FIELDS) {
        close $fh or carp $ERRNO;
        if (_has_zero_sequences(\@temp)) {
            foreach my $item (@{$thresholds}) { push @results, 0; }
            return \@results;
	} else {
	    croak "$file_path does not have expected structure";
	}
    }

    my $pf_bases = $temp[2];
    if (!check_int($pf_bases) || $pf_bases < 0 ) {
        close $fh or croak $ERRNO;
        croak("Cannot get a total number of bases from $file_path");
    }

    # go to third line
    $line = <$fh>;
    if ($line) {
        $line = <$fh>;
    }
    if (!$line) {
        close $fh or croak $ERRNO;
        croak "$file_path does not have expected structure";
    }

    @temp = split /\s+/xms, $line;
    my $num_qs = @temp - $NUM_OTHER_FIELDS;

    $line = <$fh>;
    close $fh or carp $ERRNO;

    if (!$line || $line !~ /Total/smx) {
        croak "$file_path does not have expected structure";
    }

    @temp = split /\s+/xms, $line;
    my $end = @temp - 2;
    foreach my $threshold (@{$thresholds}) {
        my $result = 0;
        if (check_int($threshold) && $threshold >= 0 && $threshold < $num_qs) {
            my $start = $POSITION_OF_QVAL_LOWEST + ($threshold - $LOWEST_VALUE);
            for my $i ($start .. $end) {
	        $result += $temp[$i];
            }
            $result = $pf_bases*$result/$COEFF;
        }
        push @results, round($result);
    }

    return \@results;
}

1;

__END__


=head1 NAME

npg_common::qXvalues

=head1 VERSION

$LastChangedRevision: 12048 $

=head1 SYNOPSIS
  This module calculates Q values by looking at a *.fastqcheck file, not at the *.fastq file.
  It is known that these values are 1-2% lower than the values calculated by parsing a *.fastq
  file. The execution is quick, the time does not depend on the size of the *.fastq file from
  which the *.fastqcheck file was derived.

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 check_int - check whether the argument is an integer, returns 1 if it is and 0 if it is not.
  if(!check_int($my_arg)) {croak qq[Fail integer test for my_arg: $my_arg];}

=head2 get_qvals - computes QX (Q20, Q30, etc) values at and above the levels given in the argument list. Returns a reference to a list of values. The values are reported in bases.
  @thresholds = qw(20 30 35);
  my $values = get_qvals('path_to_fastqcheck_file', \@thresholds);

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item strict

=item warnings

=item Carp

=item English -no_match_vars

=item Readonly

=item Exporter

=item Math::Round qw(round);

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: mg8 $

=head1 LICENSE AND COPYRIGHT

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

