#########
# Author:        Marina Gourtovaia
# Maintainer:    $Author: gq1 $
# Created:       June 2010
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: map2lims.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/roles/run/lane/map2lims.pm $
#

package npg_common::roles::run::lane::map2lims;

use strict;
use warnings;
use Moose::Role;
use Carp;
use English qw(-no_match_vars);
use Readonly;

use npg::api::run_lane;
use st::api::sample;
use st::api::asset;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd ProhibitParensWithBuiltins RequireCheckingReturnValueOfEval)

=head1 NAME

npg_common::roles::run::lane::map2lims

=head1 VERSION

$Revision: 12048 $

=head1 SYNOPSIS

=head1 DESCRIPTION

Interface (Moose role) that, given infornation about a lane,
retrievs LIMS object that are needed for specific reasons,
for example, for finding a reference. 
Requires the following method present in the implementing class: id_run, position.

=head1 SUBROUTINES/METHODS

=cut


sub _check_prerequisites {
  my $self = shift;
  eval {
    $self->id_run;
    $self->position;
    1;
  } or do {
    croak q[id_run and position attributes should be defined for the object that uses functions from npg_common::roles::run::lane::map2lims ];
  };
  return;
}

sub _get_run_lane {
  my $self = shift;

  my $hash_ref = {id_run => $self->id_run, position => $self->position,};
  if ( $self->can( q{has_api_util} ) && $self->has_api_util() ) {
    $hash_ref->{util} = $self->api_util();
  }
  my $run_lane = npg::api::run_lane->new( $hash_ref );

  if (!$run_lane) {
    croak q[run_lane object undefined for run ] . $self->id_run . q[, lane ] . $self->position;
  }
  return $run_lane;
}


=head2 asset4expected_insert_size

LIMS asset object (LibraryTube) for finding expected insert size

=cut
sub asset4expected_insert_size {
  my $self = shift;

  $self->_check_prerequisites();
  my $id_run = $self->id_run;
  my $position = $self->position;
  my $tag_index = undef;
  eval {
    $tag_index = $self->tag_index;
  };

  my $run_lane = $self->_get_run_lane();
  my $asset;
  if ($run_lane->pool && $tag_index) {
    my $st_lane = $run_lane->st_lane;
    if (!$st_lane->pool_description) {
      croak qq[No pool description for run $id_run position $position];
    }
    if (exists $st_lane->pool_description->{$tag_index}->{lib_tube_asset_id}) {
      $asset = st::api::asset->new( {id => $st_lane->pool_description->{$tag_index}->{lib_tube_asset_id}} );
    } else {
      croak qq[Library tube asset for run $id_run position $position tag index $tag_index not found];
    }
  } else {
    $asset = $run_lane->entity;
  }

  return $asset;
}


=head2 assets4ref

A reference to a list of LIMS assets that are needed to find a genome reference
for a lane or, if a tag_index attribute is defined, for an indexed component
of a pool.

given a parameter 'library', this method will return a list
only including the corresponding library object st::api::asset

=cut
sub assets4ref { ## no critic (Subroutines::ProhibitExcessComplexity)

  my ($self, $library_sample) = @_;

  $self->_check_prerequisites();
  $library_sample ||= q{sample};

  my $id_run = $self->id_run;
  my $position = $self->position;
  my $tag_index = undef;
  eval {
    $tag_index = $self->tag_index;
  };

  my $run_lane = $self->_get_run_lane();

  my @objects_to_examine = ();
  my $asset = $run_lane->entity;
  if ($run_lane->pool) {
    if ($tag_index or ($library_sample eq q{sample})) {
      my $lane = $run_lane->st_lane;
      if (!$lane->pool_description) {
	      croak qq[No pool description for run $id_run, lane $position];
      }
      my @tag_indices = $tag_index ? ($tag_index) : keys %{$lane->pool_description};
      for my $tag_index_l (@tag_indices){
        my $error_end =  qq[for tag $tag_index_l in run $id_run lane  $position];
        if (!exists $lane->pool_description->{$tag_index_l}) {
	      croak qq[No description $error_end];
        }
        my $id_name = q{sample_id};
        if($library_sample eq q{library}){
          $id_name = q{lib_tube_asset_id}
        }
        if (exists  $lane->pool_description->{$tag_index_l}->{$id_name} &&
             $lane->pool_description->{$tag_index_l}->{$id_name}) {
          my $id = $lane->pool_description->{$tag_index_l}->{$id_name};
          my $hash_st_ref = {'id' => $id,};
          if ( $self->can( q{has_api_util} ) && $self->has_api_util() ) {
            $hash_st_ref->{util} = $self->api_util();
          }
          my $class_name = q{sample};
          if($library_sample eq q{library}){
            $class_name= q{asset};
          }
          push @objects_to_examine, qq{st::api::$class_name}->new( $hash_st_ref );
        }
      }
    }elsif($library_sample eq q{library}){
       push @objects_to_examine, $asset;
    }
  } else {
    if (defined $tag_index) {
      carp qq[The caller expects a pool (tag_index $tag_index attr defined), but lane $position (run $id_run) is not a pool];
    }
    my $sample = $asset->sample;
    if ($run_lane->library && $sample && $library_sample eq q{sample}) {
        push @objects_to_examine, $sample;
    } else {
      push @objects_to_examine, $asset;
    }
  }

  return \@objects_to_examine;
}

no Moose::Role;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item strict

=item warnings

=item Moose::Role

=item Carp

=item English

=item Readonly

=item npg::api::run_lane

=item st::api::sample

=item st::api::asset

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Author: Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
