#############
# Created By: Marina Gourtovaia
# Maintainer: $Author: gq1 $
# Created On: 23 April 2010
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# $Id: file_finder.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/run/file_finder.pm $

package npg_common::run::file_finder;

use Moose;
use MooseX::StrictConstructor;
use Carp;
use English qw{-no_match_vars};
use File::Spec::Functions qw(catfile);

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12048 $ =~ /(\d+)/mxs; $r; };

extends 'npg_common::run::finder';

with    qw/ npg_common::roles::run::lane
            npg_common::roles::run::lane::tag
            npg_common::roles::run::lane::file_names
          /;


Readonly::Scalar our $FILE_EXTENSION      => q[fastq];
Readonly::Scalar our $TAGS_FILE_FLAG      => 0;
Readonly::Scalar our $LANE_ARCHIVE_LOOKUP => 0;


has 'file_extension'   =>   (
                             isa            => 'Maybe[Str]',
                             is             => 'rw',
                             required       => 0,
                             default        => $FILE_EXTENSION,
		            );

has 'with_t_file'      =>   (
                             isa            => 'Bool',
                             is             => 'rw',
                             required       => 0,
                             default        => $TAGS_FILE_FLAG,
		            );

has 'lane_archive_lookup' =>   (
                                isa            => 'Bool',
                                is             => 'rw',
                                required       => 0,
                                default        => $LANE_ARCHIVE_LOOKUP,
		               );

has '_current_archive_folder' =>  (
                                   isa            => 'Maybe[Str]',
                                   is             => 'rw',
                                   required       => 0,
		                  );

has '_globbed'         =>   (
                             isa            => 'Maybe[HashRef]',
                             is             => 'rw',
                             required       => 0,
		            );

sub BUILD {
    my $self = shift;
    $self->_test_options_compatibility();
    return;
}


sub _test_options_compatibility {
    my $self = shift;
    if (defined $self->tag_index && $self->with_t_file) {
        croak 'tag_index and with_t_file attributes cannot be both set';
    }
}

sub _append_tag_index {
    my ($self, $fname) = @_;

    my $tag_label = $self->tag_label;
    if (!$tag_label) {
        return $fname;
    }

    my $extension =  $self->file_ext($self->file_extension);
    if (!$extension) {
        $fname .= $tag_label;
    } else {
        my $delim = q[.];
        my @name_components = split /\./smx, $fname;
        pop @name_components;
        $fname= join $delim, @name_components;
        $fname .= $tag_label;
        $fname .= $extension;;
    }
    return $fname;
}

sub files { ##no critic (ProhibitExcessComplexity)

    my ($self) = @_;

    $self->_test_options_compatibility();

    my $tag_index;
    my $lane_archive_exists = 0;
    my $lane_archive = defined $self->tag_index && $self->lane_archive_lookup ?
         File::Spec->catfile($self->archive_folder, $self->lane_archive) : q[];
    if ($lane_archive) {
        if (!-e $lane_archive) {
            $tag_index = $self->tag_index;
            $self->tag_index(undef);
	} else {
            $lane_archive_exists = 1;
	}
    }

    my $archive_dir = $lane_archive_exists ? $lane_archive : $self->archive_folder;

    if (!$self->_current_archive_folder || $self->_current_archive_folder ne $archive_dir) {
        $self->_current_archive_folder($archive_dir);
        my $globbed = {};
        opendir(my $dh, $archive_dir) || croak "can't opendir $archive_dir: $ERRNO";
        my @all = readdir $dh;
        closedir $dh;
        foreach my $f (@all) {
            $globbed->{$f} = 1;
        }
        $self->_globbed($globbed);
    }

    my $fnames = {};

    foreach my $f (@{$self->generate_filename($self->file_extension)}) {
        if (exists $self->_globbed->{$f}) {
            $fnames->{forward} =  $f;
            last;
	}
    }

    if (!exists $fnames->{forward}) {
        my $forward  =  $self->generate_filename($self->file_extension, 1)->[0];
        if (exists $self->_globbed->{$forward}) {
            $fnames->{forward} =  $forward;
        }
        my $reverse =  $self->generate_filename($self->file_extension, 2)->[0];
        if (exists $self->_globbed->{$reverse}) {
	    $fnames->{reverse} =  $reverse;
        }
    }

    if ($self->with_t_file) {
        my $tag = $self->generate_filename($self->file_extension, q[t])->[0];
        if (exists $self->_globbed->{$tag}) {
             $fnames->{tags} =  $tag;
        }
    }

    if (defined $tag_index) {
        $self->tag_index($tag_index);
        $archive_dir = $lane_archive;
    }
    foreach my $type (keys %{$fnames}) {
        if (defined $tag_index) {
            $fnames->{$type} = $self->_append_tag_index($fnames->{$type});
	}
        $fnames->{$type} = File::Spec->catfile($archive_dir, $fnames->{$type});
    }

    return $fnames;
}


no Moose;
__PACKAGE__->meta->make_immutable;

1;

__END__

=head1 NAME

npg_common::run::file_finder

=head1 VERSION

$LastChangedRevision: 12048 $

=head1 SYNOPSIS

  my $finder = npg_common::run::file_finder->new(id_run => 2222, position => 1);

=head1 DESCRIPTION

Locates the files for a lane.

=head1 SUBROUTINES/METHODS

=head2 file_extension - a writeble attribute, defaults to fastq, can be empty

=head2 with_t_file - a writable boolean attribute, defaults to 0, determins whether a file for tags will be looked up

=head2 files - Returns a reference to a hash containing paths to a forward and reverse(if any) input files and
also a file for a tag if requested. The possible keys are forward, reverse, and tags.

  my $hash = $finder->files;
  my $forward = $hash->{forward};

=head2 BUILD - some sanity checking before returning a reference to a newly created object to the caller; for internal use only

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::StrictConstructor

=item Carp

=item English -no_match_vars

=item Readonly

=item npg_common::run::finder

=item npg_common::roles::run::lane

=item npg_common::roles::run::lane::tag

=item npg_common::roles::run::lane::file_names

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: gq1 $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
