#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2010-11-17
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: SAM_Qseq.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/sequence/SAM_Qseq.pm $
#

package npg_common::sequence::SAM_Qseq;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);

with 'MooseX::Getopt';

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/smx; $r; };

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::SAM_Qseq

=head1 VERSION

$LastChangedRevision: 12048 $

=head1 SYNOPSIS

=head1 DESCRIPTION

Given a sam file or a stream from standard input, add extra fields for each read

for example, E2 or U2 from qseq files of second calls

or OQ from qseq files of original basecalling

=head1 SUBROUTINES/METHODS

=head2 qseq_files_1

qseq files in array for read 1

=cut

has 'qseq_files_1'               => (isa            => 'ArrayRef',
                                     is             => 'rw',
                                     documentation  => 'qseq files for read 1',
                                     lazy_build     => 1,
                                  );

sub _build_qseq_files_1 {
  my $self = shift;

  my $read_number_1 = $self->read_number_1();
  if(! defined $read_number_1){
     return [];
  }
  my $qseq_files_1 = $self->_glob_qseq_files($read_number_1);
  if($qseq_files_1){
    return $qseq_files_1;
  }
  return [];
}

=head2 qseq_files_2

qseq files in array for read 2

=cut

has 'qseq_files_2'               => (isa            => 'ArrayRef',
                                     is             => 'rw',
                                     documentation  => 'qseq files for read 2',
                                     lazy_build     => 1,
                                  );

sub _build_qseq_files_2 {
  my $self = shift;

  my $read_number_2 = $self->read_number_2();
  if(! defined $read_number_2){
     return [];
  }
  my $qseq_files_2 = $self->_glob_qseq_files($read_number_2);
  if($qseq_files_2){
    return $qseq_files_2;
  }
  return [];
}

sub _glob_qseq_files{
  my ($self, $end) = @_;

  my $qseq_dir = $self->qseq_dir();
  my $lane = $self->lane();

  if($qseq_dir && $lane && $end){
     opendir my $qseq_dir_fh, $qseq_dir or croak "can not read dir $qseq_dir";
     my $qseq_file_pattern = q{s_}.$lane.q{_}.$end.q{_\d+_qseq[.]txt};
     my @qseq_files = map { $qseq_dir.qq{/$_}; } grep {/^$qseq_file_pattern$/mxs } readdir $qseq_dir_fh ;
     closedir $qseq_dir_fh;
     return \ @qseq_files;
  }
  return;
}

=head2 qseq_dir

qseq file directory

=cut

has 'qseq_dir'                 => (isa            => 'Str',
                                   is             => 'rw',
                                   documentation  => 'second call qseq files dir',
                                  );
=head2 lane

lane number to glob the qseq files for the given sam 

=cut

has 'lane'                     => ( isa            => 'Int',
                                    is             => 'rw',
                                    documentation  => 'lane number',
                                  );
=head2 read_number_1

read number 1 to glob the qseq files for the given sam 

=cut

has 'read_number_1'            => ( isa            => 'Int',
                                    is             => 'rw',
                                    documentation  => 'read number 1',
                                  );
=head2 read_number_2

read number 2 to glob the qseq files for the given sam 

=cut

has 'read_number_2'            => ( isa            => 'Int',
                                    is             => 'rw',
                                    documentation  => 'read number 2',
                                  );
=head2 sam

input sam file to add extra fields from qseq files
 
=cut

has 'sam'             => (isa             => 'Str',
                          is              => 'rw',
                          default         => q{},
                          documentation   => 'input sam file to add things from qseq files',
                         );

=head2 process

read qseq files line by line and add bases or scores to the sam file

=cut

sub process{ ##no critic (Subroutines::ProhibitExcessComplexity)
   my $self = shift;

   #get qseq files
   my $qseq_files_1 = $self->qseq_files_1();
   my $qseq_files_2 = $self->qseq_files_2();

   my $num_qseq_files_1 = scalar @{$qseq_files_1};
   my $num_qseq_files_2 = scalar @{$qseq_files_2};

   if(!$num_qseq_files_1 ){
     croak 'There is no second call qseq files available!';
   }

   if( $num_qseq_files_2 && ($num_qseq_files_1 != $num_qseq_files_2) ){
     croak 'The number of qseq files for read 2 is not equal with read 1';
   }

   #check where to get sam file and pre-process sam file
   my $sam_file = $self->sam();

   my $sam_fh;
   my $sam_line;

   #sam from a file or stdin
   if($sam_file){
     open $sam_fh, q{<}, $sam_file or croak "Can not open file:$sam_file"; ## no critic (InputOutput::RequireBriefOpen)
   }else{
     $sam_fh =q{STDIN};
   }
   #jump header section of sam
   $sam_line = <$sam_fh>;
   while($sam_line && $sam_line =~ /^@/mxs){
         print $sam_line or croak 'Can not print';
         $sam_line = <$sam_fh>;
   }

   my @sorted_qseq_files_2 = sort @{$qseq_files_2};

   my $count = 0 ;
   foreach my $qseq_file (sort @{$qseq_files_1}){

     open my $qseq_fh, q{<}, $qseq_file or croak "Can not open file: $qseq_file"; ## no critic (InputOutput::RequireBriefOpen)

     my $qseq_fh_2;
     my $qseq_file_2;
     if ($num_qseq_files_2){
       $qseq_file_2 = $sorted_qseq_files_2[$count];
       open $qseq_fh_2, q{<}, $qseq_file_2 or croak "Can not open file: $qseq_file_2"; ## no critic (InputOutput::RequireBriefOpen)
     }

     while(my $qseq_line = <$qseq_fh>){
        chomp $qseq_line;
        my @fields = split /\t/mxs, $qseq_line;
        my $filter = pop @fields;

        my $id = $self->_get_id_from_qseq_line(\@fields);

        if((! $filter) || !$sam_line){
          if($num_qseq_files_2){
            <$qseq_fh_2>;
          }
          next;
        }elsif($sam_line){
            chomp $sam_line;
            my @sam_fields = split /\t/mxs, $sam_line;
            my $id_sam = shift @sam_fields;
            if(($id_sam !~ /^$id/mxs ) ){
               if($num_qseq_files_2){
                  <$qseq_fh_2>;
               }
               next;
            }
        }

        print $self->_add_qseq_line($sam_line, \@fields), "\n" or croak 'can not print';
        $sam_line = <$sam_fh>;

        if($num_qseq_files_2){
           my $qseq_line_2 = <$qseq_fh_2>;
           chomp $qseq_line_2;
           my @fields_2 = split /\t/mxs, $qseq_line_2;
           pop @fields_2;

           print $self->_add_qseq_line($sam_line, \@fields_2), "\n" or croak 'can not print';
           $sam_line = <$sam_fh>;
        }
     }
     close $qseq_fh or croak "Can not close file: $qseq_file";
     if($qseq_fh_2){
       close $qseq_fh_2 or croak "Can not close file: $qseq_file_2";
     }
     $count++;
   }

   if($sam_line){
     croak "The sam file has more reads than the qseq files: $sam_line";
   }

   if($sam_fh){
     close $sam_fh or croak "Can not close file: $sam_file";
   }
   return;
}

=head2 _add_qseq_line

add second call to one line in sam

=cut

sub _add_qseq_line {
   my ($self, $sam_line, $qseq_fields) = @_;

   my $id = $self->_get_id_from_qseq_line($qseq_fields);

   my $end = $qseq_fields->[7]; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   if($end == 3){ ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
     $end = 2;
   }

   chomp $sam_line;
   my @sam_fields = split /\t/mxs, $sam_line;
   my $id_sam = shift @sam_fields;
   if($id_sam !~ /^$id/mxs){
     croak "Read id in sam file $id_sam not match id in qseq file $id!";
   }

   my $flag = shift @sam_fields;
   my $first_read_sam = ($flag & 64) >> 6; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   my $second_read_sam = ($flag & 128) >> 7; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   my $end_sam;
   if($first_read_sam){
      $end_sam = 1;
   }
   if($second_read_sam){
     $end_sam = 2;
   }

   if($end_sam && $end_sam != $end){
      croak "End number in sam file $end_sam not match end number in qseq file $end!";
   }

   my $strand_query = ( $flag & 16 ) >> 4; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)

   my $new_sam_line = $sam_line;

   my $fields_to_add = $self->_fields_to_add($qseq_fields);

   foreach my $sam_tag_name (keys %{$fields_to_add}){

      my $sam_tag = $fields_to_add->{$sam_tag_name};
      my $sam_tag_value = $sam_tag->{value};
      my $complement  = $sam_tag->{complement};
      if($strand_query && $complement){
         $sam_tag_value = $self->_reverse_complement($sam_tag_value);
      }elsif ($strand_query && !$complement ){
         $sam_tag_value = reverse $sam_tag_value;
      }
      $new_sam_line .= qq{\t$sam_tag_name:Z:$sam_tag_value}
   }

   return $new_sam_line;
}

sub _fields_to_add {
   my ( $self, $qseq_fields ) = @_;

   my $fields_to_add = {};

   return $fields_to_add;
}

sub _get_id_from_qseq_line{
  my ($self, $qseq_fields) = @_;

  my @qseq_fields_copy = @{$qseq_fields};

  my $machine = shift @qseq_fields_copy;
  my @id_fields = splice @qseq_fields_copy, 0, 5; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
  my $id = join q{:}, @id_fields;
  $id = $machine.q{_}.$id;

  return $id;
}

sub _reverse_complement {
  my ($self, $read) = @_;

  my $revese_read = reverse $read;

  $revese_read  =~ tr/atcgATCG/tagcTAGC/;

  return $revese_read;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
