#########
# Author:        Marina Gourtovaia
# Maintainer:    $Author: gq1 $
# Created:       14 April 2009
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: find.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_common/sequence/reference/roles/find.pm $
#

package npg_common::sequence::reference::roles::find;

use strict;
use warnings;
use Moose::Role;
use Carp;
use English qw(-no_match_vars);
use File::Spec::Functions qw(catfile);
use Cwd qw(abs_path);

use npg_common::messages;

with qw/ npg_common::sequence::reference::roles::list 
         npg_common::roles::run::lane::map2lims 
       /;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/smx; $r; };
## no critic (Documentation::RequirePodAtEnd ProhibitParensWithBuiltins RequireCheckingReturnValueOfEval)

Readonly::Scalar our $NPG_DEFAULT_ALIGNER_OPTION => q{npg_default};

=head1 NAME

npg_common::sequence::reference::roles::find

=head1 VERSION

$Revision: 12048 $

=head1 SYNOPSIS

An example of a class that implements this role

 package reference_user;
 use Moose;
 with qw(npg_common::sequence::reference::roles::find);

 sub id_run   { return 1937; }
 sub position { return 1; }
 sub my_function { do something;}

 1;
 __END__

Using your class

 my $r_user = reference_user->new();
 $r_user->refs();
 my @messages = $r_user->messages->messages;

See npg_qc::autoqc::checks::insert_size for an example of using this role.

=head1 DESCRIPTION

Interface (Moose role) for retrieving a reference sequence for a lane or
a list of samples.

=head1 SUBROUTINES/METHODS

=cut


Readonly::Scalar our $ALIGNER          => q[bwa];
Readonly::Scalar our $STRAIN           => q[default];
Readonly::Scalar our $SUBSET           => q[all];

Readonly::Scalar our $SEARCH_FUZZY                => 1;
Readonly::Scalar our $SEARCH_FUZZY_PHIX           => 4;
Readonly::Scalar our $SEARCH_STRICT               => 2;
Readonly::Scalar our $SEARCH_STRICT_AND_TAXON_ID  => 3;

Readonly::Scalar our $PHIX             => q[PhiX];


Readonly::Scalar our $ORGANISM_SUBSTRING_MIN_LENGTH   => 4;
Readonly::Array  our @ORGANISM_RELATED_SAMPLE_FIELDS  => qw/common_name organism public_name name/;


=head2 reference_genome

Reference genome string in the format 'organism (strain)' as set in LIMS.
See npg_common::sequence::reference::roles::list::short_report

=cut
has 'reference_genome'  => (isa             => 'Maybe[Str]',
                            is              => 'ro',
                            required        => 0,
                           );

=head2 species

Species name

=cut
has 'species'  => (isa             => 'Maybe[Str]',
                   is              => 'ro',
                   required        => 0,
                   writer          => '_set_species',
                   lazy_build      => 1,
                  );
sub _build_species {
    my $self = shift;
    my @a = $self->_parse_reference_genome;
    if (@a) {
        $self->_set_strain($a[1]);
        return $a[0];
    }
    return;
}

=head2 strain

Strain, defaults to the default strain in the repository

=cut
has 'strain'=>    (isa             => 'Str',
                   is              => 'ro',
                   required        => 0,
                   writer          => '_set_strain',
                   lazy_build      => 1,
                  );
sub _build_strain {
    my $self = shift;
    my @a = $self->_parse_reference_genome;
    if (@a) {
        $self->_set_species($a[0]);
        return $a[1];
    }
    return $STRAIN;
}

=head2 subset

Subset (ie, crhomosome), defaults to all

=cut
has 'subset'=>    (isa             => 'Str',
                   is              => 'ro',
                   required        => 0,
                   default         => $SUBSET,
                  );

=head2 search_type

A flag informing the algorithm how hard it should try to infer the species name.
The default value is $npg_common::sequence::reference::roles::find::SEARCH_FUZZY_PHIX.
Other values are $npg_common::sequence::reference::roles::find::SEARCH_STRICT_AND_TAXON_ID
 (the reference_genome and taxon_id fields of LIMS objects are examined),
 $npg_common::sequence::reference::roles::find::SEARCH_STRICT (only 
the reference_genome fields of LIMS objects are examined), and $npg_common::sequence::reference::roles::find::SEARCH_FUZZY - an algorithm that, having unsuccessfully gone through strict search types, goes into fuzzy matching of asset, sample and study names. The default search type applies the fuzzy step to Phix samples and libraries only.

Regardsless of the value of the flag, if the species or the reference_genome attribute of
this role is set, the LIMS objects will not be examined at all.

=cut
has 'search_type'  =>  (isa             => 'Int',
                        is              => 'ro',
                        required        => 0,
                        writer          => '_set_search_type',
                        default         => $SEARCH_FUZZY_PHIX,
                       );



=head2 aligner

Aligner name, defaults to bwa

=cut
has 'aligner'  => (isa             => 'Str',
                   is              => 'ro',
                   required        => 0,
                   default         => $ALIGNER,
                  );

=head2 messages

An npg_common::messages object to log the messages

=cut
has 'messages' => (isa          => 'npg_common::messages',
                   is           => 'ro',
                   required     => 0,
                   default      => sub { npg_common::messages->new(); },
                  );


=head2 refs

A reference to a list of reference paths.
If no reference found, an empty list is returned.
Examine the messages attribute after calling this function.

Optionally, takes two additional arguments. The first is an 
hash reference that, on exit, will contain a mapping of ids of
assets to reference path, undefined value for a path if no 
reference is found. The second optional argument is a reference
to a list of assets for which the genome reference s should be
found.

If the optional assets array ref is not provided, the object
consuming this role should have id_run and position fields defined.

=cut
sub refs {

  my ($self, $ref_map, $assets) = @_;

   my @refs = ();
  if ($self->species) {
    push @refs, $self->_get_reference_path($self->species, $self->strain);
  } else {

    if (!$assets) {
      $assets = $self->assets4ref;
    }

    if ( $assets && @{$assets}) {
      my $ref_hash;
      foreach my $object (@{$assets}) {
        my $path = $self->asset2ref($object);
        # abs_path for undef or empty string returns a current working directory
        if ($path) {
          $path = abs_path($path);
          $ref_hash->{$path} = 1;
        }
        if ($ref_map) {
	  $ref_map->{$object->id} = $path ? $path : undef;
        }
      }
      if ($ref_hash) {
        @refs = keys %{$ref_hash};
      }
    }
  }

  return \@refs;
}


=head2 single_ref_found

Returns true if only one reference has been found.
Returns false if no references found or multiple references found.

=cut
sub single_ref_found {

  my $self = shift;
  carp 'This method is depricated. Please use the refs method and evaluate the size of the returned array.';

  my @refs;
  eval {
    @refs = @{$self->refs()};
    1;
  } or do {
    return 0;
  };
  if (!@refs || scalar @refs > 1) { return 0; }
  return 1;
}


=head2 reset_strain

Reset the strain value to default.

=cut
sub reset_strain {

  my $self = shift;
  $self->_set_strain($STRAIN);
  return;
}


=head2 ref_info

An npg_common::sequence::reference::info object or undef

=cut
sub ref_info {
  my $self = shift;
  if (scalar @{$self->refs} == 1) {

    my $refpath = $self->refs->[0];
    my $ref = npg_common::sequence::reference::info->new(aligner => $self->aligner, ref_path => $refpath);
    my $opts_in = $ref->ref_path . q[.options];
    if (-e $opts_in) {
      ## no critic (RequireBriefOpen)
      open my $fh, q[<], $opts_in or croak qq[Cannot open $opts_in for reading: $ERRNO];
      my $found = q{};
      while (my $line = <$fh>) {
        if ($line ne qq[\n] && $line !~ /^\#/smx) {
          $line =~ s/^\s+//sxm;
	       $line =~ s/\s+$//sxm;
	       $found = $line;
	       last;
	     }
      }
      close $fh or croak qq[Cannot close $opts_in : $ERRNO];
      ## use critic

	   $ref->aligner_options($found);

    }else{
       $ref->aligner_options($NPG_DEFAULT_ALIGNER_OPTION);
    }
    return $ref;
  }elsif(scalar @{$self->refs} > 1) {

     carp 'More than one reference found: '. (join q[;], @{$self->refs});
     return;
  }

  carp 'No reference found';
  return;
}


=head2 _get_reference_path

Returns a path to a binary reference (with a prefix of the reference itself). The organism should be one of the listed in our repository.

=cut
sub _get_reference_path {

  my ($self, $organism, $strain) = @_;

  if (!$organism) {
      croak q[Organism should be defined in arguments to _get_reference_path()];
  }

  $strain = $strain || $self->strain;

  # check that the directory for the chosen aligner exists
  my $base_dir = catfile($self->ref_repository, $organism, $strain, $self->subset);
  my $dir = catfile($base_dir, $self->aligner);

  if (!-e $dir) {
    ##no critic (ProhibitInterpolationOfLiterals)
    my $message = sprintf("Binary %s reference for %s, %s, %s does not exist; path tried %s",
			  $self->aligner, $organism, $strain, $self->subset, $dir);
    ##use critic
    croak $message;
  }

  # read the fasta directory and get the file name with the reference
  my $fasta_dir = catfile($base_dir, q[fasta]);
  opendir my $dh, $fasta_dir or croak $ERRNO;
  my @files = readdir $dh;
  closedir $dh or croak $ERRNO;

  for my $file (@files) {
    if ($file =~ /\.fn?a$/ismx || $file =~ /\.fasta$/ismx) {
      # add the directory separator at the end of the aligner directory path
      # and then append the reference file name, which is the prefix of the
      # binary reference file - that is our convention
      return catfile($dir, $file);
    }
  }
  croak qq[Reference file with .fa or .fasta or .fna extension not found in $fasta_dir];
}

=head2 _map_organism_name

Converts a free text string with an organism name to the name of the organism that can be used as input to _get_reference_path. If the conversion fails, an undefined value is returned.

 my $organism = $r_user->_map_organism_name(q[clostridium difficile]);
 print $organism; #prints 'Clostridium_difficile'

=cut
sub _map_organism_name {

  my ($self, $name) = @_;

  if (!$name) {return;}

  my $count = 0;
  my $match = q[];
  my $delim = $self->organism_name_delim;

  foreach my $known_organism (@{$self->organisms}) {
      my @known_words = split /$delim/smx, $known_organism;
      my $index = (scalar @known_words) - 1 ;
      while ($index >= 0) {
	my $known_word = $known_words[$index];
        if (length $known_word >= $ORGANISM_SUBSTRING_MIN_LENGTH) {
          if ($name =~ /$known_word/ismx) {
	    if ($index != 0) {
              return $known_organism;
	    } else {
	      $count++;
              $match = $known_organism;
	    }
          }
        }
        $index = $index -1;
      }
  }
  if ($match) {
    if ($count > 1) {
      $self->messages->push(qq[Genus in $name matches to several organisms]);
    } else {
      return $match;
    }
  }

  return;
}


=head2 asset2ref

Returns a path to a binary reference (with a prefix of the reference itself) for an asset object.
Undefined value returned if the search has failed

=cut
sub asset2ref { ## no critic (ProhibitExcessComplexity)
  my ($self, $asset) = @_;

  # All kinds of asset object can be passed to this method.
  # We cannot know what fields are defined for a particular object type,
  # hence we use eval when accessing the fields/methods of the object.

  # Give preference to a reference defined by the user either in a sample
  # or study objects (reference_genome accessor)
  my $ref_path = $self->_get_preset_reference_path($asset);
  if ($ref_path || $self->search_type == $SEARCH_STRICT) {
    return $ref_path;
  }

  my @names = ();

  # Try taxon id field
  my $taxon_id;
  eval { $taxon_id = $asset->taxon_id; };
  if ($taxon_id) {
    my $description = $self->taxonid2species($taxon_id);
    if ($description->{species})  {
      my $strain =  $description->{strain} ?  $description->{strain} : q[];
      $ref_path = $self->_get_reference_path($description->{species}, $strain);
    }
    if (!$ref_path) { $self->messages->push(qq[no reference for taxon id $taxon_id]); }
  }

  if ($ref_path || $self->search_type == $SEARCH_STRICT_AND_TAXON_ID) {
    return $ref_path;
  }

  # Then, if unseccessful with taxon id, try organism, common_name, public_name and name field
  # Return from the loop early if found
  foreach my $field (@ORGANISM_RELATED_SAMPLE_FIELDS) {
    my $name;
    eval { $name = $asset->$field; };
    if ($name) {
      push @names, $name;
      my $organism = $self->_map_organism_name($name);
      if ($organism) {
        $ref_path = $self->_get_reference_path($organism);
        if ($ref_path) {
	  if($self->search_type == $SEARCH_FUZZY_PHIX && $ref_path !~ /$PHIX/smx) {
	    $ref_path = q[];
	  }
          last;
        }
      }
    }
  }

  if ($ref_path || $self->search_type == $SEARCH_FUZZY_PHIX) {
    return $ref_path;
  }

  # And, finally, if all of the above failed, try to interpret the project name
  my $study;
  eval { $study = $asset->study; };
  if ($study) {
    my $study_name = $study->name;
    my $organism = $self->_study_name2organism($study_name);
    if ($organism) {
      $ref_path = $self->_get_reference_path($organism);
    }
    if ($ref_path) {
      $self->messages->push('Organism derived from study name');
    } else {
      push @names, $study_name;
    }
  }

  if (!$ref_path) {
    $self->messages->push(q[Failed to fuzzy match ] . join(q[;], @names) . q[ to known organism names]);
  }

  return $ref_path;
}


sub _get_preset_reference_path {
  my ($self, $asset) = @_;

  if (!$asset) {
      croak q[Asset not defined];
  }

  my $preset_genome ||= $self->reference_genome;

  my $type = ref $asset;
  my $object;
  if ($type !~ /st::api::sample/smx) {
    eval { $object = $asset->sample(); };
  } else {
    $object = $asset;
  }

  if (!$preset_genome && $object) {
    $preset_genome = $object->reference_genome;
    if (!$preset_genome) {
      my $study;
      # We get only one study here
      # TODO change when ->studies is available
      eval { $study = $object->study(); };
      if ($study) {
	$object = $study;
        $preset_genome = $object->reference_genome;
      }
    }
  }

  if ($preset_genome) {
    return $self->_preset_ref2ref_path($preset_genome, $object);
  }
  return q[];
}


sub _parse_reference_genome {
  my ($self, $reference_genome) = @_;
  $reference_genome ||= $self->reference_genome;
  if ($reference_genome) {
    my @a = $reference_genome =~/(\S+)\s+\((\S+)\)/smx;
    if (scalar @a >= 2 && $a[0] && $a[1]) {
      return @a;
    }
  }
  return;
}


sub _preset_ref2ref_path {
  my ($self, $ref, $object) = @_;

  if (!$ref) {
      croak q[Reference genome is not defined or empty];
  }

  my $ref_path = q[];
  if ($ref eq $npg_common::sequence::reference::roles::list::NO_ALIGNMENT_OPTION) {
    $self->messages->push($ref);
    if ($self->search_type != $SEARCH_STRICT) {
      $self->_set_search_type($SEARCH_STRICT);
    }
  } else {
    my $error = q[];
    if ($object) {
      $error = q[in object ] . ref($object) . q[ with id ] . $object->id;
    }
    my ($species, $strain) = $self->_parse_reference_genome($ref);
    if ($species && $strain) {
      $ref_path = $self->_get_reference_path($species, $strain);
    } else {
      $self->messages->push(qq[Incorrect reference genome format $ref $error]);
    }
  }
  return $ref_path;
}


sub _study_name2organism {

  my ($self, $project_name) = @_;
  my $organism;
  if ( $project_name =~ /1000.*genomes|human/ismx || $project_name =~ /CGP/smx ) {
    $organism = q[Human];
  } elsif ( $project_name =~ /mouse|mice/ismx ) {
    $organism = q[Mouse];
  } else {
    $organism = q[];
  }
  return $organism;
}

no Moose::Role;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item strict

=item warnings

=item Moose::Role

=item Carp

=item English

=item File::Spec::Functions

=item Readonly

=item Cwd

=item npg_common::messages

=item npg_common::roles::run::lane::map2lims

=item npg_common::sequence::reference::roles::list

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Author: Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
