#########
# Author:        mg8
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $ $Author: gq1 $
# Id:            $Id: db.pm 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/lib/npg_testing/db.pm $
#

package npg_testing::db;

use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};
use YAML qw(LoadFile DumpFile);
use File::Temp qw(tempfile);
use File::Spec::Functions qw(catfile);
use Moose::Role;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12048 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd ProhibitParensWithBuiltins)

Readonly::Scalar our $FEATURE_EXTENSION => q[.yml];

=head1 NAME

npg_testing::db

=head1 VERSION

$Revision: 12048 $

=head1 SYNOPSIS

=head1 DESCRIPTION

A Moose role for creating and loading a test sqlite database using an existing DBIx database binding

=head1 SUBROUTINES/METHODS

=cut


=head2 rs_list2fixture

Dumps a list of result sets for a table to a YAML fixture.
Takes three arguments: (1) the name of the result class name for a table,
(2) a reference to a list with result sets all coming from the same table,
(3) a path to dump the fixture to.

=cut
sub rs_list2fixture {
    my ($self, $tname, $rs_list, $path) = @_;
    my @rows = ();
    foreach my $rs (@{$rs_list}) {
        while (my $r = $rs->next) {
            push @rows, {$r->get_columns};
        }
    }
    DumpFile(catfile($path,$tname).$FEATURE_EXTENSION, \@rows);
    return;
}


=head2 load_fixtures

Loads YAML fixtures to the test database. The fixtures files should be
named TableName.yml, where 'TableName' is the name of the relevant
table as known to the DBIx database binding.

=cut
sub load_fixtures {

    my ($self, $schema, $path) = @_;
    opendir my $dh, $path or croak "Could not open $path";
    my @fixtures = grep { /.yml$/smix } readdir $dh;
    if (scalar @fixtures == 0) { croak qq[no fixtures found at $path]; }

    for my $fx (@fixtures) {
      my $yml     = LoadFile("$path/$fx");
      my @temp = split /\./sxm, $fx;
      pop @temp;
      my $table = join q[.], @temp;
      carp "+- Loading $fx into $table";
      foreach my $row (@{$yml}) {
	$schema->resultset($table)->create($row);
      }
    }
    return;
}


=head2 create_test_db

Creates a temp test database and loads data into it.
The first argument is a DBIx Schema object full namespace, 
the second is the path to the directory where the fixtures
are located, the third one is a file name for the temporary
SQLite database. If the third argument is not given, a randomly
named file in a temporary directory is used; this file will
be cleaned up on exit.

=cut
sub create_test_db {
    my ($self, $schema_package, $fixtures_path, $tmpdbfilename) = @_;

    if (!$schema_package) { croak q[Schema package undefined in create_test_db]; }
    if (!$fixtures_path)  { croak q[Fixtures path undefined in create_test_db]; }

    ##no critic (ProhibitStringyEval RequireCheckingReturnValueOfEval)
    eval "require $schema_package" or do { croak $EVAL_ERROR;} ;
    ##use critic

    if (!defined $tmpdbfilename) {
       my $fh;
       ($fh,$tmpdbfilename) = tempfile(UNLINK => 1);
    }

    my $tmpschema = $schema_package->connect('dbi:SQLite:'.$tmpdbfilename);
    $tmpschema->deploy;
    $self->load_fixtures($tmpschema,  $fixtures_path);
    return $tmpschema;
}


no Moose::Role;
1;

__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item warnings

=item strict

=item Moose::Role

=item Carp

=item English

=item YAML

=item File::Temp

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Author: Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut


