#########
# Author:        Marina Gourtovaia
# Maintainer:    $Author: gq1 $
# Created:       29 July 2009
# Last Modified: $Date: 2010-12-02 12:30:28 +0000 (Thu, 02 Dec 2010) $
# Id:            $Id: 10-sequence-reference-find.t 12048 2010-12-02 12:30:28Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-31.0/t/10-sequence-reference-find.t $
#

package reference;

use strict;
use warnings;
use Test::More tests => 77;
use Test::Exception;
use File::Spec::Functions qw(catfile);
use Cwd qw(cwd);
use Moose::Meta::Class;

my $repos = catfile(cwd, q[t/data/sequence/references]);
my $bwa_human_ref = q[Human/NCBI36/all/bwa/someref.fa];

use_ok('npg_common::sequence::reference::roles::find');

{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,
                          strain => q[some-strain],
                       });
  throws_ok { $ruser->_get_reference_path() } qr/Organism\ should\ be\ defined/, 
           'croak on organism not defined';
  throws_ok { $ruser->_get_reference_path(q[PhiX]) } qr/Binary bwa reference for PhiX, some-strain, all does not exist/, 'error message when strain is not available';

  $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,
                         aligner => q[some],
                       });
  throws_ok { $ruser->_get_reference_path(q[PhiX]) } qr/Binary some reference for PhiX, default, all does not exist/, 'error message when aligner does not exist';
  throws_ok { $ruser->_get_reference_path(q[PhiX], q[my_strain]) } qr/Binary some reference for PhiX, my_strain, all does not exist/, 'error message when aligner does not exist';

  $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,
                          subset => q[chr3]
                       });
  throws_ok { $ruser->_get_reference_path(q[PhiX]) } qr/Binary bwa reference for PhiX, default, chr3 does not exist/, 'error message for non-existing subset';
  throws_ok { $ruser->_get_reference_path(q[human]) } qr/Binary\ bwa\ reference/, 'error message when the directory structure for the binary ref is missing';

  $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,});
  is ($ruser->_get_reference_path(q[Human]), catfile($repos, q[Human/default/all/bwa/someref.fa]), 
           'correct reference path'); 
  throws_ok { $ruser->_get_reference_path(q[Human], q[no_ref_strain]) } qr/Reference file with .fa or .fasta or .fna extension not found in/, 'error message when no genome ref is found in the fasta directory';
  is ($ruser->_get_reference_path(q[Human], q[fna_strain]), catfile($repos, q[Human/fna_strain/all/bwa/someref.fna]), 'genome reference with .fna extension is found');

  $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, aligner => 'fasta'});
  is ($ruser->_get_reference_path(q[Human], q[fna_strain]), catfile($repos, q[Human/fna_strain/all/fasta/someref.fna]), 'genome reference with .fna extension is found');
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,});
  my $new_dir = catfile($repos, q[Fox]);
  my $new_slink = catfile($repos, q[my_reddy]);
  if (!-e $new_dir) { mkdir $new_dir; }

  eval { symlink($new_dir, $new_slink); };

  SKIP: {
    if (!-e $new_slink) {
      skip 'Should be possible to create a symbolic link', 3;
    }
    is(join(q[ ], sort(@{$ruser->organisms()})), q[Bordetella_bronchiseptica Clostridium_difficile Fox Human PhiX Streptococcus_equi Streptococcus_pneumoniae my_reddy], 'known organisms retrieved from the file system');
    ok(!$ruser->_map_organism_name(q[my]), 'failed to map organism name for my');
    ok(!$ruser->_map_organism_name(q[dog_my]), 'failed to map organism name for dog_my');
  }

  if (-e $new_slink) {unlink $new_slink};
  rmdir $new_dir;
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,});
  my $new_dir = catfile(cwd, $repos, q[Fox]);
  if (!-e $new_dir) { mkdir $new_dir; }

  ok(!$ruser->_map_organism_name(q[Fox]), 'overly short organism names are skipped');
  rmdir $new_dir;
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,});

  is($ruser->_map_organism_name(q[clostridium difficile]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[C difficile]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[C. difficile]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[C.difficile]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile Clostridium]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile,Clostridium]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile_Clostridium]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile.Clostridium]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile, Clostridium]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[difficile C.]), 'Clostridium_difficile', 'organism mapping');
  is($ruser->_map_organism_name(q[HUMAN]), 'Human', 'organism mapping');
  is($ruser->_map_organism_name(q[phix_illumina]), 'PhiX', 'phix_illumina maps to PhiX');
  is($ruser->_map_organism_name(q[phixillumina]), 'PhiX', 'phixillumina maps to PhiX');
  is($ruser->_map_organism_name(q[illuminaphix]), 'PhiX', 'illuminaphix maps to PhiX');
  ok(!$ruser->_map_organism_name(q[any]), 'failed to map organism name for any');
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,
                         organisms => ['Bordetella_bronchiseptica', 'Bordetella_pertussis', 'homo_sapiens'],
                       });

  is ($ruser->_map_organism_name(q[Bordetella Bronchiseptica]), q[Bordetella_bronchiseptica], 'match on species OK');
  is ($ruser->_map_organism_name(q[Bordetella pertussis]), q[Bordetella_pertussis], 'match on species OK');
  is ($ruser->_map_organism_name(q[homo]), q[homo_sapiens], 'match on genus OK when only 1 species is available');
  is ($ruser->messages->count, 0, 'no messages');
  is ($ruser->_map_organism_name(q[my Bordetella]), undef, 'match on genus not OK when 2 species in a genus are available');
  is ($ruser->messages->count, 1, 'one error message');
  is ($ruser->messages->pop, 'Genus in my Bordetella matches to several organisms', 'genus only identification failed error');
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos,});
  is ( $ruser->_study_name2organism('1000Genomes-B1-JPT'), 'Human', 'reference for 1000 genoms project');
  is ( $ruser->_study_name2organism('1000-genomes project'), 'Human', 'reference for 1000 genoms project');
  is ( $ruser->_study_name2organism('my CGP project'), 'Human',  'reference for CGP project');
  ok ( !$ruser->_study_name2organism('mycgp project'), 'no reference for cgp project');
  is ( $ruser->_study_name2organism('human evolution'), 'Human', 'reference for human evolution project');
  is ( $ruser->_study_name2organism('my mouse project'), 'Mouse',    'reference for a mouse project');
  is ( $ruser->_study_name2organism('my MICE project'), 'Mouse',    'reference for mice project');
}


{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, });
  throws_ok { $ruser->_preset_ref2ref_path() } qr/Reference genome is not defined or empty/, 'missing or empty attribute  error';
  throws_ok { $ruser->_preset_ref2ref_path(q[]) } qr/Reference genome is not defined or empty/, 'missing or empty attribute  error';

  is ($ruser->_preset_ref2ref_path(q[Tiger]), q[], 'incorrect ref genome - return empty string');
  like ($ruser->messages->pop, qr/Incorrect reference genome format Tiger/, 'incorrect ref genome format error logged');

  is ($ruser->_preset_ref2ref_path(q[Tiger ()]), q[], 'incorrect ref genome - return empty string');
  like ($ruser->messages->pop, qr/Incorrect reference genome format Tiger ()/, 'incorrect ref genome format error logged');

  is ($ruser->_preset_ref2ref_path(q[ (tiger)]), q[], 'incorrect ref genome - return empty string');
  like ($ruser->messages->pop, qr/Incorrect reference genome format/, 'incorrect ref genome format error logged');

  is ($ruser->_preset_ref2ref_path(q[ ]), q[], 'incorrect ref genome - return empty string');
  like ($ruser->messages->pop, qr/Incorrect reference genome format  /, 'incorrect ref genome format error logged');

  throws_ok {$ruser->_preset_ref2ref_path(q[Human (dada) ])} qr/Binary bwa reference for Human, dada, all does not exist/, 'non-existing strain error';
  throws_ok {$ruser->_preset_ref2ref_path(q[dodo (fna_strain) ])} qr/Binary bwa reference for dodo, fna_strain, all does not exist/, 'non-existing organism error';
}

{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, });

  is($ruser->strain, q[default], 'default strain');
  is($ruser->species, undef, 'species not defined by default');
  is($ruser->reference_genome, undef, 'reference_genome not defined by default');
}

{
  my $species = q[Human];
  my $strain = q[NCBI36];
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, species => $species, strain => $strain});

  is($ruser->strain, $strain, qq[strain $strain]);
  is($ruser->species, $species, qq[species $species]);
  is($ruser->reference_genome, undef, 'reference_genome not defined by default');
  is($ruser->refs->[0], catfile($repos, $bwa_human_ref), 'path to human reference');
}

{
  my $species = q[Human];
  my $strain = q[NCBI36];
  my $reference_genome = $species . q[ (] . $strain .q[)];
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, reference_genome => $reference_genome});

  is($ruser->strain, $strain, qq[strain $strain]);
  is($ruser->species, $species, qq[species $species]);
  is($ruser->reference_genome, $reference_genome, qq[reference_genome $reference_genome]);
  is($ruser->refs->[0], catfile($repos, $bwa_human_ref), 'path to human reference');
}

{
  my $species = q[Human];
  my $strain = q[NCBI36];
  my $reference_genome = q[dodo];
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos, 
                         species => $species, 
                         strain => $strain, 
                         reference_genome => $reference_genome});

  is($ruser->strain, $strain, qq[strain $strain]);
  is($ruser->species, $species, qq[species $species]);
  is($ruser->reference_genome, $reference_genome, qq[reference_genome $reference_genome]);
  is($ruser->refs->[0], catfile($repos, $bwa_human_ref), 'path to human reference');
}

{
  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos });
  
  is(join(q[ ], $ruser->_parse_reference_genome(q[Salmonella_enterica (Enteritidis_P125109)])), q[Salmonella_enterica Enteritidis_P125109], 'ref genome parsing');
  is(join(q[ ], $ruser->_parse_reference_genome(q[Homo_sapiens (CGP_GRCh37.NCBI.allchr_MT)])), q[Homo_sapiens CGP_GRCh37.NCBI.allchr_MT], 'ref genome parsing');
  is($ruser->_parse_reference_genome(q[Salmonella_enterica]), undef, 'ref genome parsing');
}

{
  require npg_common::sequence::reference::roles::list;
  my $no_align = $npg_common::sequence::reference::roles::list::NO_ALIGNMENT_OPTION;

  my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({ ref_repository => $repos });
  is($ruser-> _preset_ref2ref_path($no_align), q[], 'no preset ref for no-align option');
  is($ruser->search_type, $npg_common::sequence::reference::roles::find::SEARCH_STRICT, 'search mode reset correctly');
  is($ruser->messages->pop, $no_align, 'correct message saved');
}